package at.asit.project.timing;

import android.content.ContextWrapper;
import android.util.Log;
import android.util.Pair;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import org.jetbrains.annotations.Contract;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.CountDownLatch;

import at.asit.project.timing.parsedTypes.APIClassInfo;
import at.asit.project.timing.parsedTypes.APIMethodInfo;
import at.asit.project.timing.types.APIClass;
import at.asit.project.timing.types.APIMethod;
import at.asit.project.timing.types.CustomHashMap;

public class APIClassLoader {
  private static final boolean LOGGING = false;
  private final String TAG = APIClassLoader.class.getName();

  private final CustomHashMap<String, APIClass> methodsOfClasses = new CustomHashMap<>();
  private final ContextWrapper contextWrapper;
  private final Predefines predefines;
  private final Parser parser;
  private final CustomHashMap<String, APIClassInfo> methodsOfClassesToCheckParsed; // methods to invoke in the first exploration step (from Config.[methods|constructors]_file_resource)
  private final CustomHashMap<String, APIClassInfo> methodsOfAllClassesParsed; // for parameter creation and return value exploration (from Config.all_[methods|constructors]_file_resource)

  public APIClassLoader(ContextWrapper contextWrapper, Predefines predefines, @NonNull Parser parser) throws IOException {
    this.contextWrapper = contextWrapper;
    this.predefines = predefines;
    this.parser = parser;
    methodsOfClassesToCheckParsed = parser.getMethodsOfClassesToCheckParsed();
    methodsOfAllClassesParsed = parser.getMethodsOfAllClassesParsed();

    loadPredefinedClasses(contextWrapper);
    loadParsedClasses();

  }

  public CustomHashMap<String, APIClass> getMethodsOfClasses() {
    return methodsOfClasses;
  }

  // exploration phase
  private void saveInvokableMethods(Class<?> clazz, ArrayList<Object> classObjects, int currentRecursionDepth, String currentKnownClass, Object parentClassObject, Method objectCreatorMethod, Object[] objectCreatorMethodParameters,
                                    CustomHashMap<String, APIClassInfo> methodsOfClassesParsed, String hierarchy) throws IOException {
    for (Method methodToInvoke : clazz.getDeclaredMethods()) {
      boolean addedMethod = false;
      methodToInvoke.setAccessible(true);
      APIClassInfo apiClassInfo = methodsOfClassesParsed.getOrDefault(currentKnownClass, null);

      boolean blacklisted = Config.isInBlacklist(methodToInvoke.getName(), clazz.getName());

      if (methodToInvoke.getName().contains("getArtManager")) {
        continue;
      }

      if (Config.isRelevantMethod(methodToInvoke.getName()) && !blacklisted && (!Config.USE_ONLY_METHODS_FROM_FILE || (apiClassInfo != null && apiClassInfo.methodExists(methodToInvoke.getName())))) {
        if (LOGGING) {
          Log.v(TAG, "Checking methodToInvoke: " + methodToInvoke.getName() + " class object size: " + classObjects.size() + " on " + clazz.getName());
        }

        ArrayList<Object> classObjectsAdapted = new ArrayList<>(classObjects);
        if (Modifier.isStatic(methodToInvoke.getModifiers()) && Modifier.isPublic(methodToInvoke.getModifiers())) {
          classObjectsAdapted.add(null);
        }
        for (Object classObject : classObjectsAdapted) {
          if (LOGGING) {
            Log.v(TAG, "is static");
            Log.v(TAG, "clazz.getDeclaredMethods() = " + Arrays.toString(clazz.getDeclaredMethods()));
            Log.v(TAG, "methodToInvoke = " + methodToInvoke.toString());
          }

          ArrayList<ArrayList<Object>> parameterObjectsList = createMethodParameterObjects(currentKnownClass, ((classObject != null) ? classObject.getClass().getName() : currentKnownClass), methodToInvoke);

          for (ArrayList<Object> parameterObjectsArrayList : parameterObjectsList) {
            Object[] parameterObjects = parameterObjectsArrayList.toArray();
            String packageAndMethod = clazz.toString() + ":" + methodToInvoke.toString();
            String newHierarchy = hierarchy + "||" + packageAndMethod;
            try {
              if (LOGGING) {
                Log.v(TAG, "parameterObjects = " + Arrays.toString(parameterObjects) + " of method " + methodToInvoke.getName());
              }


              MethodReturn methodReturn = invokeMethod(methodToInvoke, classObject, parameterObjects, 10);
              Object returnValue = methodReturn.getActualReturnValue();
              MethodReturn methodReturnRedone = null;

              if (methodReturn.getTimings() != null) {
                MethodReturn.AveragedReturnTimes averagedReturnTimes = methodReturn.getAverageTiming();
                System.out.println("averagedReturnTimes = " + averagedReturnTimes + "for method " + methodToInvoke.toString());

                // redo measurements if percentage above certain point with more runs, e.g. if above 10% difference
                if (averagedReturnTimes.percentage > .1 || averagedReturnTimes.percentage < -.1) {
                  methodReturnRedone = invokeMethod(methodToInvoke, classObject, parameterObjects, 100);
                  System.out.println("methodReturnRedone = " + methodReturnRedone);
                }
              }

              if (LOGGING) {
                //Log.i(TAG, "Method calls took " + Arrays.toString(diff));
                Log.v(TAG, "returnValue = " + returnValue + " of method: " + methodToInvoke.getName());
                Log.v(TAG, "currentRecursionDepth = " + currentRecursionDepth + " newHierarchy = " + newHierarchy);
              }
              if (returnValue != null && !Utils.isWrapperType(returnValue.getClass()) && !Utils.isCollectionType(returnValue) && !(currentRecursionDepth >= Config.EXPLORATION_DEPTH_FOR_CLASS_RETURN_VALUES && (!Config.USE_ONLY_METHODS_FROM_FILE))) {
                if (LOGGING) {
                  Log.v(TAG, "returnValue.getTypeName() = " + Utils.getTypeName(returnValue.getClass()) + " of method " + methodToInvoke.getName() + "() of class " + clazz.getName() + " of parentclass: " + ((parentClassObject != null) ?
                          parentClassObject.getClass().getName() : "null") + "(method: " + ((objectCreatorMethod != null) ? objectCreatorMethod.getName() : "null") + "())");
                }

                String superType = null;

                if (currentRecursionDepth < Config.EXPLORATION_DEPTH_FOR_CLASS_RETURN_VALUES && (!Config.USE_ONLY_METHODS_FROM_FILE || superType != null)) {
                  ArrayList<Object> classObjectsNext = new ArrayList<>();
                  classObjectsNext.add(returnValue);
                  if (LOGGING) {
                    Log.v(TAG, "GOING INTO THE DEEP FOR returnValue: " + returnValue.getClass().getName());
                  }
                  // explore return value recursively
                  saveInvokableMethods(returnValue.getClass(), classObjectsNext, currentRecursionDepth + 1, superType, classObject, methodToInvoke, parameterObjects, methodsOfAllClassesParsed, newHierarchy);
                }
              }
//else {
              APIClass apiClass = methodsOfClasses.get(clazz.getName());
              if (apiClass == null) {
                apiClass = new APIClass(currentKnownClass, classObject);
              }

              APIMethod apiMethod = apiClass.addMethod(contextWrapper.getApplicationContext(), classObject, methodToInvoke, parameterObjects, newHierarchy);

              String fileContent = null;
              if (methodReturnRedone != null && methodReturnRedone.getTimings() != null) {
                fileContent = methodReturnRedone.toString();
                if (!fileContent.equals(apiMethod.previouslyWrittenToFile) || (!Config.FILTER_SAME_VALUES && APIMethod.KEEP_MEASUREMENTS)) {
                  apiMethod.previouslyWrittenToFile = fileContent;
                  apiMethod.addMeasurement(fileContent, true);
                }
              } else if (methodReturn.getTimings() != null) {
                fileContent = methodReturn.toString();
                if (!fileContent.equals(apiMethod.previouslyWrittenToFile) || (!Config.FILTER_SAME_VALUES && APIMethod.KEEP_MEASUREMENTS)) {
                  apiMethod.previouslyWrittenToFile = fileContent;
                  apiMethod.addMeasurement(fileContent, false);
                }
              }

              // save container class for later recreation of object
              if (Config.CONTAINER_CLASSES.contains(currentKnownClass)) {
                if (LOGGING) {
                  Log.v(TAG, "Saving container class " + parentClassObject + " objectCreatorMethod: " + objectCreatorMethod + " objectCreatorMethodParameters: " + Arrays.toString(objectCreatorMethodParameters));
                }
                apiMethod.setParentCreators(parentClassObject, objectCreatorMethod, objectCreatorMethodParameters);
              }

              addedMethod = true;
            } catch (Exception e) {
              e.printStackTrace();
              throw new RuntimeException(e);
            }
          }
        }
      }
      if (LOGGING) {
        Log.v(TAG, (addedMethod ? "[+?] Added (for now)" : "[-] Removed") + " method " + methodToInvoke.getName());
      }
    }
  }

  private boolean containsCallback(@NonNull Object[] parameterObjects) {
    for (Object obj : parameterObjects) {
      if (obj instanceof Proxy) {
        return true;
      }
    }
    return false;
  }

  @NonNull
  @Contract("_, _, _, _ -> new")
  private MethodReturn invokeMethod(@NonNull Method method, Object classObject, Object[] parameterObjects, int rounds) throws InvocationTargetException, IllegalAccessException, InterruptedException {
    long[] diffs = new long[rounds];
    long startTime = 0L, endTime = 0L;
    boolean exception = false;
    Pair<Boolean, Object[]> adaptedParams = getAdaptedParameterValues(parameterObjects);
    Object[] adaptedParameterObjects = adaptedParams.second;
    Object returnValue = null;
    int firstPart = rounds / 2;

    startTime = System.nanoTime();
    try {
      returnValue = method.invoke(classObject, parameterObjects);
    } catch (Exception e) {
      exception = true;
    }
    endTime = System.nanoTime();
    diffs[0] = endTime - startTime;
    if (!adaptedParams.first) {
      return new MethodReturn(returnValue, null, parameterObjects, adaptedParameterObjects);
    }

    if (containsCallback(parameterObjects)) {
      invokeMethodsRepeatedlyCallback(method, classObject, parameterObjects, rounds, diffs, exception, adaptedParameterObjects, firstPart);
    } else {
      invokeMethodsRepeatedly(method, classObject, parameterObjects, rounds, diffs, exception, adaptedParameterObjects, firstPart);
    }

    return new MethodReturn(returnValue, diffs, parameterObjects, adaptedParameterObjects);
  }

  private void invokeMethodsRepeatedly(@NonNull Method method, Object classObject, Object[] parameterObjects, int rounds, long[] diffs, boolean exception, Object[] adaptedParameterObjects, int firstPart) throws IllegalAccessException,
          InvocationTargetException {
    long startTime;
    long endTime;
    if (exception) {
      startTime = System.nanoTime();
      try {
        method.invoke(classObject, parameterObjects);
      } catch (Exception ignored) {
      }
      endTime = System.nanoTime();
      diffs[0] = endTime - startTime;

      for (int i = firstPart; i < rounds; ++i) {
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, adaptedParameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }

      for (int i = firstPart; i < rounds; ++i) {
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, adaptedParameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }

      for (int i = 0; i < firstPart; ++i) {
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, parameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }
    } else {
      startTime = System.nanoTime();
      method.invoke(classObject, parameterObjects);
      endTime = System.nanoTime();
      diffs[0] = endTime - startTime;

      for (int i = firstPart; i < rounds; ++i) {
        startTime = System.nanoTime();
        method.invoke(classObject, adaptedParameterObjects);
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }

      for (int i = firstPart; i < rounds; ++i) {
        startTime = System.nanoTime();
        method.invoke(classObject, adaptedParameterObjects);
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }

      for (int i = 0; i < firstPart; ++i) {
        startTime = System.nanoTime();
        method.invoke(classObject, parameterObjects);
        endTime = System.nanoTime();
        diffs[i] = endTime - startTime;
      }
    }
  }

  private void invokeMethodsRepeatedlyCallback(@NonNull Method method, Object classObject, Object[] parameterObjects, int rounds, long[] diffs, boolean exception, Object[] adaptedParameterObjects, int firstPart) throws IllegalAccessException,
          InvocationTargetException, InterruptedException {
    long startTime = 0L;
    long endTime = 0L;
    if (exception) {
      CallbackListener.countDownLatch = new CountDownLatch(1);
      startTime = System.nanoTime();
      try {
        method.invoke(classObject, parameterObjects);
      } catch (Exception ignored) {
      }
      endTime = System.nanoTime();
      CallbackListener.countDownLatch.await();
      diffs[0] = CallbackListener.endTime - startTime;

      for (int i = firstPart; i < 10; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, adaptedParameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
        System.out.println("continuing");
      }

      for (int i = firstPart; i < rounds; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, adaptedParameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
      }

      for (int i = 0; i < firstPart; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        try {
          method.invoke(classObject, parameterObjects);
        } catch (Exception ignored) {
        }
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
      }
    } else {
      CallbackListener.countDownLatch = new CountDownLatch(1);
      startTime = System.nanoTime();
      method.invoke(classObject, parameterObjects);
      endTime = System.nanoTime();
      CallbackListener.countDownLatch.await();
      diffs[0] = CallbackListener.endTime - startTime;

      for (int i = firstPart; i < rounds; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        method.invoke(classObject, adaptedParameterObjects);
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
      }

      for (int i = firstPart; i < rounds; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        method.invoke(classObject, adaptedParameterObjects);
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
      }

      for (int i = 0; i < firstPart; ++i) {
        CallbackListener.countDownLatch = new CountDownLatch(1);
        startTime = System.nanoTime();
        method.invoke(classObject, parameterObjects);
        endTime = System.nanoTime();
        CallbackListener.countDownLatch.await();
        diffs[i] = CallbackListener.endTime - startTime;
      }
    }
  }

  @NonNull
  private Pair<Boolean, Object[]> getAdaptedParameterValues(@NonNull Object[] originalParameterValues) {
    Random random = new Random();
    boolean adapted = false;
    Object[] adaptedParameterValues = new Object[originalParameterValues.length];
    for (int counter = 0; counter < originalParameterValues.length; ++counter) {
      Object originalValue = originalParameterValues[counter];
      if (originalValue == null) {
        adaptedParameterValues[counter] = null;
        continue;
      }
      System.out.println("originalValue = " + originalValue.getClass());
      if (originalValue instanceof String) {
        adaptedParameterValues[counter] = new String(new char[((String) originalValue).length()]).replace('\0', 'a');
        adapted = true;
      } else if (originalValue instanceof Integer) {
        adaptedParameterValues[counter] = random.nextInt();
        adapted = true;
      } else if (originalValue instanceof Character) {
        adaptedParameterValues[counter] = random.nextInt(26) + 'a';
        adapted = true;
      } else if (originalValue instanceof Byte) {
        adaptedParameterValues[counter] = (byte) random.nextInt();
        adapted = true;
      } else if (originalValue instanceof Short) {
        adaptedParameterValues[counter] = (short) random.nextInt();
      } else if (originalValue instanceof Double) {
        adaptedParameterValues[counter] = random.nextDouble();
        adapted = true;
      } else if (originalValue instanceof Long) {
        adaptedParameterValues[counter] = random.nextLong();
        adapted = true;
      } else if (originalValue instanceof Float) {
        adaptedParameterValues[counter] = random.nextFloat();
        adapted = true;
      } else if (originalValue instanceof Void) {
        throw new RuntimeException();
      } else if (originalValue instanceof Boolean) {
        adaptedParameterValues[counter] = !((Boolean) originalValue);
        adapted = true;
      } else if (Utils.isCollectionType(originalValue)) {
        if (originalValue.getClass().isArray()) {
          Object[] originalValues = ((Object[]) originalValue);
          Object[] newValues = new Object[originalValues.length];
          for (int valueCtr = 0; valueCtr < originalValues.length; ++valueCtr) {
            Pair<Boolean, Object[]> returnValue = getAdaptedParameterValues(new Object[]{originalValues[valueCtr]});
            if (returnValue.first) {
              adapted = true;
            }
            newValues[valueCtr] = returnValue.second[0];
          }
          adaptedParameterValues[counter] = Utils.castArray(newValues);
        } else {
          throw new RuntimeException("Not implemented");
        }
      } else {
        adaptedParameterValues[counter] = originalParameterValues[counter];
      }
    }
    return new Pair<>(adapted, adaptedParameterValues);
  }

  private void parameterObjectPermutations(@NonNull ArrayList<ArrayList<Object>> parameterObjects, ArrayList<Object> permutation, ArrayList<ArrayList<Object>> result, int position) {
    if (position < parameterObjects.size()) {
      for (int i = 0; i < parameterObjects.get(position).size(); ++i) {
        permutation.add(parameterObjects.get(position).get(i));
        parameterObjectPermutations(parameterObjects, permutation, result, position + 1);
        permutation.remove(permutation.size() - 1);
      }
    } else {
      result.add(new ArrayList<>(permutation));
    }
  }

  @NonNull
  private ArrayList<ArrayList<Object>> createMethodParameterObjects(String knownClassNameOfMethod, String realClassName, Method method) throws IOException {
    ArrayList<String[]> allParameterNames = new ArrayList<>();
    if (LOGGING) {
      Log.v(TAG, "trying to get params. class: " + knownClassNameOfMethod + " method.getName() = " + method.getName() + " private: " + Modifier.isPrivate(method.getModifiers()) + " protected: " + Modifier
              .isProtected(method.getModifiers()) + " public: " + Modifier.isPublic(method.getModifiers()));
    }

    APIClassInfo apiClassInfo = methodsOfAllClassesParsed.getOrDefault(knownClassNameOfMethod, null);
    if (apiClassInfo != null && apiClassInfo.methodExists(method.getName())) {
      allParameterNames = apiClassInfo.getMethodInfo(method.getName()).getParameterNamesWithLength(method.getParameterTypes().length);
    }

    // Write interface param names to file
    int index = 0;
    for (Class<?> parameterType : method.getParameterTypes()) {
      if (parameterType.isInterface()) {
        try {
          StringBuilder interfaceParameters = new StringBuilder(parameterType.toString() + ":::");
          for (String[] parameterNames : allParameterNames) {
            interfaceParameters.append("{").append(parameterNames[index]).append("} ");
          }
        } catch (IndexOutOfBoundsException e) {
          e.printStackTrace();
        }
      }
      ++index;
    }

    //init array
    ArrayList<ArrayList<Object>> parameterObjects = new ArrayList<>();
    for (int paramCounter = 0; paramCounter < method.getParameterTypes().length; ++paramCounter) {
      parameterObjects.add(new ArrayList<>());
    }

    for (int nameCounter = 0; nameCounter < allParameterNames.size(); ++nameCounter) {
      String[] parameterNames = allParameterNames.get(nameCounter);
      for (int paramCounter = 0; paramCounter < method.getParameterTypes().length; ++paramCounter) {
        Object predefinedValue = predefines.getPredefinedParameterValue(knownClassNameOfMethod, realClassName, method.getName(), method.toString(), parameterNames[paramCounter], method.getParameterTypes()[paramCounter].getName());

        if (predefinedValue instanceof LinkedHashMap) {
          LinkedHashMap<String, Object> predefinedObject = (LinkedHashMap<String, Object>) predefinedValue;
          if ((boolean) predefinedObject.get("parameter_is_array")) {
            predefinedValue = predefinedObject.get("value");

            if (Utils.isAssignable(predefinedValue.getClass(), method.getParameterTypes()[paramCounter], true)) {
              parameterObjects.get(paramCounter).add(predefinedValue);
              Log.v(TAG, "ADDED ARRAY predefinedObject = " + predefinedObject);
            } else if (predefinedValue instanceof Collection && (Utils.isAssignable(((Collection) predefinedValue).toArray().getClass(), method.getParameterTypes()[paramCounter], true) || Utils
                    .isArrayAssignable(((Collection) predefinedValue).toArray(), method.getParameterTypes()[paramCounter]))) {
              // Array handling
              Object[] arr = Utils.castArray(((Collection) predefinedValue).toArray());
              parameterObjects.get(paramCounter).add(arr);
              Log.v(TAG, "ADDED ARRAY.toArray() predefinedObject = " + predefinedObject);
            }
          } else {
            ArrayList<Object> predefinedValues = ((ArrayList<Object>) predefinedObject.get("value"));
            for (Object predefinedValue_ : predefinedValues) {
              if (predefinedValue_ != null && Utils.isAssignable(predefinedValue_.getClass(), method.getParameterTypes()[paramCounter], true)) {
                parameterObjects.get(paramCounter).add(predefinedValue_);
                Log.v(TAG, "ADDED NON ARRAY predefinedObject = " + predefinedObject);
              }
            }
          }
        } else if (!(predefinedValue instanceof Predefines.ParameterNotFound) && (predefinedValue == null || Utils.isAssignable(predefinedValue.getClass(), method.getParameterTypes()[paramCounter], true))) {
          parameterObjects.get(paramCounter).add(predefinedValue);
        }
      }
    }

    for (int paramCounter = 0; paramCounter < method.getParameterTypes().length; ++paramCounter) {
      Class<?> parameterType = method.getParameterTypes()[paramCounter];

      // Callback handling
      if (parameterType.isInterface()) {
        if (Config.isPossibleCallback(parameterType.getName(), parameterType.getMethods())) {
          Class<?>[] parameterClassArray = {parameterType};
          Object proxy = Proxy.newProxyInstance(parameterType.getClassLoader(), parameterClassArray, new CallbackListener());
          parameterObjects.get(paramCounter).add(proxy);
        }
      }

      if (parameterObjects.get(paramCounter).isEmpty()) {
        Object defaultParameter = Utils.getDefaultParameter(parameterType);
        if (defaultParameter != null) {
          parameterObjects.get(paramCounter).add(defaultParameter);
        }
      }

      if (parameterObjects.get(paramCounter).isEmpty()) {
        Constructor<?>[] constructors = parameterType.getConstructors();
        try {
          ArrayList<Object> classObjects = createClassObject(constructors, parser.getConstructorsOfAllClassesParsed());
          for (Object classObject : classObjects) {
            if (classObject != null) {
              parameterObjects.get(paramCounter).add(classObject);
            }
          }
        } catch (IllegalAccessException | InvocationTargetException | InstantiationException e) {
          e.printStackTrace();
        }
      }
    }

    ArrayList<ArrayList<Object>> result = new ArrayList<>();
    parameterObjectPermutations(parameterObjects, new ArrayList<>(), result, 0);
    return result;
  }

  private ArrayList<Object> createClassObject(Constructor[] constructors, HashMap<String, ArrayList<String[]>> constructorsAndParameters) throws IllegalAccessException, InvocationTargetException, InstantiationException {
    ArrayList<Object> constructorObjects = new ArrayList<>();

    constructorLoop:
    for (Constructor constructor : constructors) {
      ArrayList<ArrayList<Object>> parameterObjects = new ArrayList<>();
      String[] parameterNames = getConstructorParameterNames(constructorsAndParameters, constructor);

      for (int paramCounter = 0; paramCounter < constructor.getParameterTypes().length; ++paramCounter) {
        parameterObjects.add(new ArrayList<>());
      }

      for (int paramCounter = 0; paramCounter < constructor.getParameterTypes().length; ++paramCounter) {
        Class paramClass = constructor.getParameterTypes()[paramCounter];
        ArrayList<Object> classObjects = new ArrayList<>();

        Object predefinedValue = null;
        if (parameterNames != null) {
          String[] splittedConstructorName = constructor.getName().split("\\.");
          predefinedValue = predefines.getPredefinedParameterValue(constructor.getName(), constructor.getName(),
                  //TODO real name vs known name
                  splittedConstructorName[splittedConstructorName.length - 1], constructor.toString(), parameterNames[paramCounter], paramClass.getName());
        }

        if (predefinedValue instanceof LinkedHashMap) {
          LinkedHashMap<String, Object> predefinedObject = (LinkedHashMap<String, Object>) predefinedValue;
          if ((boolean) predefinedObject.get("parameter_is_array")) {
            predefinedValue = predefinedObject.get("value");
            if (Utils.isAssignable(predefinedValue.getClass(), constructor.getParameterTypes()[paramCounter], true)) {
              classObjects.add(predefinedValue);
              if (LOGGING) {
                Log.v(TAG, "ADDED ARRAY predefinedObject = " + predefinedObject);
              }
            }
          } else {
            ArrayList<Object> predefinedValues = ((ArrayList<Object>) predefinedObject.get("value"));
            for (Object predefinedValue_ : predefinedValues) {
              if (predefinedValue_ != null && Utils.isAssignable(predefinedValue_.getClass(), constructor.getParameterTypes()[paramCounter], true)) {
                //TODO null ptr
                classObjects.add(predefinedValue_);
                if (LOGGING) {
                  Log.v(TAG, "ADDED NON ARRAY predefinedObject = " + predefinedObject);
                }
              }
            }
          }
        }

        if (predefinedValue != null & paramClass.isInstance(predefinedValue)) {
          classObjects.add(predefinedValue);
        } else if (Utils.getDefaultParameter(paramClass) != null) {
          classObjects.add(Utils.getDefaultParameter(paramClass));
        } else if (Modifier.isAbstract(paramClass.getModifiers()) || paramClass.getName().equals(constructor.getName()) || (paramClass.getName().equals("android.os.Bundle") && constructor.getName()
                .equals("android.os.PersistableBundle")) || (paramClass.getName().equals("android.telecom.AudioState") && constructor.getName().equals("android.telecom.CallAudioState")) || (paramClass.getName()
                .equals("java.io.PipedInputStream") && constructor.getName().equals("java.io.PipedOutputStream")) || (paramClass.getName().equals("java.io.PipedReader") && constructor.getName().equals("java.io.PipedWriter"))) {
          continue constructorLoop;
        } else if (paramClass.isEnum()) {
          classObjects = new ArrayList<>(Arrays.asList(paramClass.getEnumConstants()));
        } else {
          classObjects = createClassObject(paramClass.getConstructors(), constructorsAndParameters);
          if (classObjects.size() < 1) {
            continue constructorLoop;
          }
        }

        parameterObjects.get(paramCounter).addAll(classObjects);
      }

      if (LOGGING) {
        Log.v(TAG, "constructor.getParameterTypes() = " + Arrays.toString(constructor.getParameterTypes()));
        if (!constructor.getName().equals("android.hardware.usb.UsbDeviceConnection") && !constructor.getName().equals("android.app.WallpaperInfo") && !constructor.getName().equals("android.mtp.MtpDevice") && !constructor.getName()
                .equals("android.view.inputmethod.InputMethodInfo")) {
          Log.v(TAG, "created parameterObjects = " + parameterObjects + " for constructor " + (!constructor.getName().equals("android.hardware.usb.UsbDeviceConnection") ? constructor.toString() : constructor.getName()));
        }
      }

      if (!constructor.getName().equals("java.math.BigInteger") && !constructor.toString().contains("public android.database.CursorWindow(java.lang.String,long)") && !constructor.getName().equals("android.media.AudioTrack") && !constructor.getName()
              .equals("android.app.Instrumentation") && !constructor.getName().equals("java.lang.Throwable")) {
        ArrayList<ArrayList<Object>> result = new ArrayList<>();
        parameterObjectPermutations(parameterObjects, new ArrayList<>(), result, 0);
        for (ArrayList<Object> parameterObject : result) {
          try {
            constructorObjects.add(constructor.newInstance(parameterObject.toArray()));
          } catch (Exception e) {
            e.printStackTrace();
          }
        }
      } else {
      }
    }
    return constructorObjects;
  }

  @Nullable
  private String[] getConstructorParameterNames(HashMap<String, ArrayList<String[]>> constructorsAndParameters, Constructor constructor) {
    ArrayList<String[]> parameterNames = constructorsAndParameters.get(constructor.getName());
    if (parameterNames != null) {
      if (LOGGING) {
        parameterNames.forEach(strings -> Log.v(TAG, "parameterNames all: " + Arrays.toString(strings)));
      }
      for (String[] possibleParameters : parameterNames) {
        if (possibleParameters.length == constructor.getParameterTypes().length) {
          System.arraycopy(possibleParameters, 0, possibleParameters, 0, possibleParameters.length);
          return possibleParameters;
        }
      }
    }
    return null;
  }

  // loads parsed classes, creates object and explores the methods by calling the saveInvokableMethods method
  private void loadParsedClasses() throws IOException {
    HashMap<String, ArrayList<String[]>> constructorsAndParameters = parser.getConstructorsOfClassesToCheckParsed();
    int notFoundClasses = 0, errorClasses = 0, totalClasses = 0;

    Log.v(TAG, "Loading the following constructorsAndParameters: " + Collections.singletonList(constructorsAndParameters));

    for (Map.Entry<String, APIClassInfo> parsedMethodsOfClass : methodsOfClassesToCheckParsed.entrySet()) {
      String currentClass = parsedMethodsOfClass.getKey();

      Class<?> clazz = null;
      ArrayList<Object> classObjects = new ArrayList<>();
      try {
        ++totalClasses;
        clazz = Class.forName(currentClass);
        Log.v(TAG, " Loading class: " + clazz.getName());

        if (Modifier.isAbstract(clazz.getModifiers()) || Modifier.isInterface(clazz.getModifiers())) {
          int missedMethodsDueToAbstractLocal = 0;
          StringBuilder methodsMissedDueToAbstractOrInterface = new StringBuilder();
          for (Map.Entry<String, APIMethodInfo> method : parsedMethodsOfClass.getValue().methodInfos.entrySet()) {
            if (Config.isRelevantMethod(method.getKey())) {
              missedMethodsDueToAbstractLocal += method.getValue().getParameterNames().size();
              methodsMissedDueToAbstractOrInterface.append(method.getKey()).append(": ");
              for (String[] paramName : method.getValue().getParameterNames()) {
                methodsMissedDueToAbstractOrInterface.append(Arrays.toString(paramName)).append(", ");
              }
              methodsMissedDueToAbstractOrInterface.append(" || ");
            }
          }
          continue;
        }
        Constructor<?>[] constructors = clazz.getConstructors();
        classObjects = createClassObject(constructors, constructorsAndParameters);
      } catch (Exception e) {
        if (e instanceof ClassNotFoundException) {
          ++notFoundClasses;
          int missedMethodsDueToNotFoundLocal = 0;
          StringBuilder methodsHere = new StringBuilder();
          for (Map.Entry<String, APIMethodInfo> method : parsedMethodsOfClass.getValue().methodInfos.entrySet()) {
            if (Config.isRelevantMethod(method.getKey())) {
              missedMethodsDueToNotFoundLocal += method.getValue().getParameterNames().size();
              methodsHere.append(method.getKey()).append(": ");
              for (String[] paramName : method.getValue().getParameterNames()) {
                methodsHere.append(Arrays.toString(paramName)).append(", ");
              }
              methodsHere.append(" || ");
            }
          }
          continue;
        } else {
          ++errorClasses;
          int missedMethodsDueToClassCreationErrorThisClass = 0;
          for (Map.Entry<String, APIMethodInfo> method : parsedMethodsOfClass.getValue().methodInfos.entrySet()) {
            if (Config.isRelevantMethod(method.getKey())) {
              missedMethodsDueToClassCreationErrorThisClass += method.getValue().getParameterNames().size();
            }
          }
        }
        e.printStackTrace();
      }
      if (clazz == null) {
        throw new RuntimeException("No class found");
        //continue;
      }

      if (classObjects.size() == 0) {
        for (Map.Entry<String, APIMethodInfo> method : parsedMethodsOfClass.getValue().methodInfos.entrySet()) {
          if (Config.isRelevantMethod(method.getKey())) {
            boolean first = true;
            int methodsCounted = 0;
            int methodsSaved = 0;
            for (Method realMethod : clazz.getDeclaredMethods()) {
              if (realMethod.getName().equals(method.getKey()) && !Modifier.isStatic(realMethod.getModifiers()) && Modifier.isPublic(realMethod.getModifiers())) {
                methodsCounted += 1;
                if (first) {
                  methodsSaved += method.getValue().getParameterNames().size();
                  first = false;
                }
              }
            }
          }
        }
      }

      int methods = 0;
      for (Map.Entry<String, APIMethodInfo> method : parsedMethodsOfClass.getValue().methodInfos.entrySet()) {
        if (Config.isRelevantMethod(method.getKey())) {
          methods += method.getValue().getParameterNames().size();
        }
      }
      saveInvokableMethods(clazz, classObjects, 0, Utils.getTypeName(clazz), null, null, null, methodsOfClassesToCheckParsed, "");
    }

    Log.i(TAG, "notFoundClasses = " + notFoundClasses + ", error classes = " + errorClasses + " of " + totalClasses);
  }

  private void loadPredefinedClasses(ContextWrapper contextWrapper) {
    HashMap<String, Object> classObjects = Predefines.loadCustomObjects(contextWrapper);
    for (Map.Entry<String, Object> classObject : classObjects.entrySet()) {
      ArrayList<Object> objectList = new ArrayList<>();
      String className = classObject.getKey();
      try {
        Class clazz = Class.forName(className);
        Object castedObject = clazz.cast(classObject.getValue());
        objectList.add(castedObject);
        saveInvokableMethods(clazz, objectList, 0, className, null, null, null, methodsOfClassesToCheckParsed, "");
      } catch (ClassNotFoundException e) {
        e.printStackTrace();
        throw new RuntimeException("Predefined class not found!");
      } catch (IOException e) {
        throw new RuntimeException();
      }
    }
  }


}
