package at.asit.project.timing;

import android.content.ContextWrapper;
import android.os.Handler;
import android.os.Looper;
import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.annotation.RawRes;
import android.util.Log;
import android.webkit.WebView;

import org.yaml.snakeyaml.Yaml;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import at.asit.project.timing.types.InterpretationResult;
import bsh.EvalError;
import bsh.Interpreter;

public class Predefines {
  private final String TAG = Predefines.class.getName();
  private final FileOutputStream exceptionsDuringEvaluation;
  private final ContextWrapper contextWrapper;
  private LinkedHashMap<String, Object> predefinedParameterValues;

  public Predefines(ContextWrapper contextWrapper) {
    this.contextWrapper = contextWrapper;
    File exceptionWhileCreationFile = new File(contextWrapper.getExternalFilesDir(null), "predefineEvaluationExceptions.info");
    try {
      exceptionsDuringEvaluation = new FileOutputStream(exceptionWhileCreationFile);
    } catch (FileNotFoundException e) {
      e.printStackTrace();
      throw new RuntimeException();
    }
    loadParameterMap(contextWrapper);
  }

  private static HashMap<String, Object> loadCustomParameterObjects(ContextWrapper contextWrapper) {
    HashMap<String, Object> customParameterObjects = new HashMap<>();
    return customParameterObjects;
  }

  static HashMap<String, Object> loadCustomObjects(ContextWrapper contextWrapper) {
    HashMap<String, Object> customObjects = new HashMap<>();
    // Add custom objects to call methods on here
    customObjects.put("android.content.Context", contextWrapper.getApplicationContext());
    Object lock = new Object();
    final WebView[] wv = new WebView[1];
    final Handler handler = new Handler(Looper.getMainLooper());
    try {
      handler.post(() -> {
        WebView w = new WebView(contextWrapper);
        synchronized (lock) {
          wv[0] = w;
          lock.notify();
        }
      });
    } catch (Exception e) {
      e.printStackTrace();
    }
    try {
      synchronized (lock) {
        while (wv[0] == null) {
          lock.wait();
        }
      }
    } catch (InterruptedException e) {
      throw new RuntimeException();
    }
    return customObjects;
  }

  private static String getMethodSignatureWithoutModifiers(String methodName, @NonNull String methodSignature, String packageAndKnownClass, String packageAndRealClass) {
    String[] methodNames = methodSignature.split(" ");
    String methodWithoutModifiers = methodNames[methodNames.length - 1];
    for (int ctr = 0; ctr < methodNames.length; ++ctr) {
      if (methodNames[ctr].contains("throws")) {
        methodWithoutModifiers = methodNames[ctr - 1];
      }
    }
    methodWithoutModifiers = methodWithoutModifiers.replace(packageAndRealClass + ".", "");
    methodWithoutModifiers = methodWithoutModifiers.replace(packageAndKnownClass + ".", "");
    if (!methodWithoutModifiers.startsWith(methodName)) {
      //throw new RuntimeException();
    }
    return methodWithoutModifiers;
  }

  private void setPredefinedVariables(Interpreter interpreter) throws EvalError {
    interpreter.set("context", contextWrapper);
  }

  private InterpretationResult interpretStatement(Object statement) {
    String strStatement = "";
    Interpreter interpreter = new Interpreter();
    try {
      setPredefinedVariables(interpreter);
      strStatement = "result = " + statement;
      return new InterpretationResult(interpreter.eval(strStatement), true);
    } catch (Error | EvalError | SecurityException evalError) {
      evalError.printStackTrace();
      try {
        exceptionsDuringEvaluation.write("-----------------------------------------------\n".getBytes());
        exceptionsDuringEvaluation.write(("Failure while evaluating: \"" + strStatement + "\" Exception:\n").getBytes());
        exceptionsDuringEvaluation.write(evalError.toString().getBytes());
        exceptionsDuringEvaluation.write("-----------------------------------------------\n".getBytes());
      } catch (IOException e) {
        e.printStackTrace();
      }
      //throw new RuntimeException(Config.LOGCAT_RUNTIME_ERROR + "Could not parse Statement \"" + statement + "\"");
      return null;
    }
  }

  private void loadConstantsParameterMap(@NonNull ContextWrapper contextWrapper, @RawRes int values, boolean isConstructor) {
    InputStream input = contextWrapper.getResources().openRawResource(values);
    Yaml yaml = new Yaml();
    LinkedHashMap<String, Object> parameterValues = (LinkedHashMap<String, Object>) yaml.load(input);
    int interpretationFail = 0, interpretationSuccess = 0;

    for (Map.Entry<String, Object> parameterStatement : parameterValues.entrySet()) {
      if (parameterStatement.getValue() instanceof String) {
        InterpretationResult interpreterResult = interpretStatement(parameterStatement.getValue());
        if (interpreterResult != null) {
          ++interpretationSuccess;
          Log.v(TAG, "interpreterResult = " + interpreterResult + " for " + parameterStatement.getKey());
          Object returnValue = predefinedParameterValues.put(parameterStatement.getKey(), interpreterResult.getResult());
          if (returnValue != null) {
            throw new RuntimeException("Value exists twice: " + parameterStatement.getKey());
          }
        } else {
          ++interpretationFail;
        }
      } else {
        LinkedHashMap<String, LinkedHashMap<String, Object>> packageAndClassMap = (LinkedHashMap<String, LinkedHashMap<String, Object>>) predefinedParameterValues.getOrDefault(parameterStatement.getKey(), new LinkedHashMap<>());
        for (Map.Entry<String, LinkedHashMap<String, Object>> methodEntry : ((LinkedHashMap<String, LinkedHashMap<String, Object>>) parameterStatement.getValue()).entrySet()) {
          if (!isConstructor && !Config.isRelevantMethod(methodEntry.getKey())) {
            continue;
          }
          LinkedHashMap<String, Object> methodInterpretedParameterMap = packageAndClassMap.getOrDefault(methodEntry.getKey(), new LinkedHashMap<>());
          for (Map.Entry<String, Object> parameterEntry : methodEntry.getValue().entrySet()) {
            Object value = parameterEntry.getValue();
            if (!(value instanceof LinkedHashMap)) { // Move single elements into array
              ArrayList<String> valueArray = new ArrayList<>();
              valueArray.add((String) value);
              value = new LinkedHashMap<String, Object>();
              ((LinkedHashMap<String, Object>) value).put("value", valueArray);
              ((LinkedHashMap<String, Object>) value).put("parameter_is_array", false);
            }
            ArrayList<String> valueArray = (ArrayList<String>) ((LinkedHashMap<String, Object>) value).get("value");
            ArrayList<Object> interpreterResults = new ArrayList<>();
            for (String possibleValue : valueArray) {
              possibleValue = possibleValue.replace("#", ".");
              InterpretationResult interpreterResult = interpretStatement(possibleValue);
              if (interpreterResult != null) {
                ++interpretationSuccess;
                interpreterResults.add(interpreterResult.getResult());
                Log.v(TAG, "interpreterResult = " + interpreterResult + " for " + parameterStatement.getKey());
              } else {
                ++interpretationFail;
              }
            }
            if (interpreterResults.size() > 0) {
              LinkedHashMap<String, Object> interpretedValues = new LinkedHashMap<>();
              interpretedValues.put("value", interpreterResults);
              interpretedValues.put("parameter_is_array", ((LinkedHashMap<String, Object>) value).get("parameter_is_array"));
              Object returnValue = methodInterpretedParameterMap.put(parameterEntry.getKey(), interpretedValues);
              if (returnValue != null) {
                throw new RuntimeException("Value exists twice: " + parameterStatement.getKey());
              }
            }
          }
          packageAndClassMap.put(methodEntry.getKey(), methodInterpretedParameterMap);
        }
        predefinedParameterValues.put(parameterStatement.getKey(), packageAndClassMap);
      }
    }

    for (Map.Entry<String, Object> customParameterObject : loadCustomParameterObjects(contextWrapper).entrySet()) {
      Object returnValue = predefinedParameterValues.put(customParameterObject.getKey(), customParameterObject.getValue());
      if (returnValue != null) {
        throw new RuntimeException("Value exists twice: " + customParameterObject.getKey());
      }
    }

    Log.v(TAG, "predefinedParameterValues = " + predefinedParameterValues);
  }

  private void loadParameterMap(@NonNull ContextWrapper contextWrapper) {
    InputStream input = contextWrapper.getResources().openRawResource(R.raw.parameter_values);
    Yaml yaml = new Yaml();
    LinkedHashMap<String, Object> parameterValues = (LinkedHashMap<String, Object>) yaml.load(input);
    LinkedHashMap<String, Object> primitiveParameterMap = (LinkedHashMap<String, Object>) parameterValues.get("primitiveParameterValues");
    LinkedHashMap<String, Object> parameterStatementsMap = (LinkedHashMap<String, Object>) parameterValues.get("parameterStatements");

    predefinedParameterValues = primitiveParameterMap;

    for (Map.Entry<String, Object> parameterStatement : parameterStatementsMap.entrySet()) {
      if (parameterStatement.getValue() instanceof String) {
        InterpretationResult interpreterResult = interpretStatement(parameterStatement.getValue());
        if (interpreterResult != null) {
          Log.v(TAG, "interpreterResult = " + interpreterResult + " for " + parameterStatement.getKey());
          Object returnValue = predefinedParameterValues.put(parameterStatement.getKey(), interpreterResult.getResult());
          if (returnValue != null) {
            throw new RuntimeException("Value exists twice: " + parameterStatement.getKey());
          }
        }
      } else {
        LinkedHashMap<String, LinkedHashMap<String, Object>> packageAndClassMap = (LinkedHashMap<String, LinkedHashMap<String, Object>>) predefinedParameterValues.getOrDefault(parameterStatement.getKey(), new LinkedHashMap<>());
        for (Map.Entry<String, LinkedHashMap<String, Object>> methodEntry : ((LinkedHashMap<String, LinkedHashMap<String, Object>>) parameterStatement.getValue()).entrySet()) {
          LinkedHashMap<String, Object> methodMap = packageAndClassMap.getOrDefault(methodEntry.getKey(), new LinkedHashMap<>());
          for (Map.Entry<String, Object> parameterEntry : methodEntry.getValue().entrySet()) {
            InterpretationResult interpreterResult = interpretStatement(parameterEntry.getValue());
            if (interpreterResult != null) {
              Log.v(TAG, "interpreterResult = " + interpreterResult + " for " + parameterStatement.getKey());
              Object returnValue = methodMap.put(parameterEntry.getKey(), interpreterResult.getResult());
              if (returnValue != null) {
                throw new RuntimeException("Value exists twice: " + parameterStatement.getKey());
              }
            }
          }
          packageAndClassMap.put(methodEntry.getKey(), methodMap);
        }
        primitiveParameterMap.put(parameterStatement.getKey(), packageAndClassMap);
      }
    }

    for (Map.Entry<String, Object> customParameterObject : loadCustomParameterObjects(contextWrapper).entrySet()) {
      Object returnValue = predefinedParameterValues.put(customParameterObject.getKey(), customParameterObject.getValue());
      if (returnValue != null) {
        throw new RuntimeException("Value exists twice: " + customParameterObject.getKey());
      }
    }

    Log.v(TAG, "predefinedParameterValues = " + predefinedParameterValues);
  }

  @Nullable
  Object getPredefinedParameterValue(String packageAndKnownClass, String packageAndRealClass, String methodName, String methodSignature, String parameterName, String parameterType) {
    if (predefinedParameterValues.containsKey(packageAndKnownClass)) {
      LinkedHashMap<String, Object> classInfos = (LinkedHashMap<String, Object>) predefinedParameterValues.get(packageAndKnownClass);
      String methodNameWithoutModifiers = getMethodSignatureWithoutModifiers(methodName, methodSignature, packageAndKnownClass.replace("." + methodName, ""), packageAndRealClass.replace("." + methodName, ""));
      if (classInfos.containsKey(methodName) && classInfos.containsKey(methodNameWithoutModifiers)) {
        throw new RuntimeException("predefined values should not be defined twice");
      }

      if (classInfos.containsKey(methodNameWithoutModifiers)) { // is a constant
        methodName = methodNameWithoutModifiers;
      }

      if (classInfos.containsKey(methodName)) {
        LinkedHashMap<String, Object> methodInfos = (LinkedHashMap<String, Object>) classInfos.get(methodName);
        if (methodInfos.containsKey(parameterName)) {
          Log.v(TAG, "found parameter: " + packageAndKnownClass + " " + methodName + " " + parameterName + " " + methodInfos.get(parameterName));
          return methodInfos.get(parameterName);
        }
      }

      if (classInfos.containsKey("*")) {
        LinkedHashMap<String, Object> methodInfos = (LinkedHashMap<String, Object>) classInfos.get("*");
        if (methodInfos.containsKey(parameterName)) {
          Log.v(TAG, "found parameter by wildcard: " + packageAndKnownClass + " " + methodName + " " + parameterName + " " + methodInfos.get(parameterName));
          return methodInfos.get(parameterName);
        }
      }
    }

    if (predefinedParameterValues.containsKey(parameterName)) {
      Log.v(TAG, "found parameter without class: " + parameterName + " " + predefinedParameterValues.get(parameterName));
      return predefinedParameterValues.get(parameterName);
    }

    if (parameterType.contains("android.content.Context")) {
      Log.v(TAG, "Found Context required but parameter not called context: " + parameterName + " " + parameterType);
      if (!predefinedParameterValues.containsKey("context")) {
        throw new RuntimeException("context not defined in predefines");
      }
      return predefinedParameterValues.get("context");
    }

    return new ParameterNotFound();
  }

  public class ParameterNotFound {
  }
}
