package at.asit.project.timing.services;

import android.app.IntentService;
import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.content.Context;
import android.content.Intent;
import android.os.Build;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.core.app.NotificationCompat;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;

import at.asit.project.timing.APIClassLoader;
import at.asit.project.timing.Config;
import at.asit.project.timing.Parser;
import at.asit.project.timing.Predefines;
import at.asit.project.timing.Utils;
import at.asit.project.timing.parsedTypes.APIClassInfo;
import at.asit.project.timing.types.APIClass;
import at.asit.project.timing.types.APIMethod;
import at.asit.project.timing.types.CustomHashMap;

public class APIHarvestingService extends IntentService {
  private static final String TAG = APIHarvestingService.class.getName();
  private static APIHarvestingService instance;

  public APIHarvestingService() {
    super("APIHarvestingService");
  }

  public static APIHarvestingService getInstance() {
    return instance;
  }

  public static void startAPIHarvestingService(Context context) {
    Intent intent = new Intent(context, APIHarvestingService.class);
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
      Log.i(TAG, "Starting foreground service");
      context.startForegroundService(intent);
    } else {
      context.startService(intent);
    }
  }

  public static void readFields(@NonNull Object classObject, String className, Class<?> class_) {
    try {
      class_ = Class.forName(className);

      Field[] fields = class_.getDeclaredFields();
      for (Field field : fields) {
        field.setAccessible(true);
        Object returnValue;
        System.out.println("Accessing field.getName() = " + field.getName());
        try {
          if (java.lang.reflect.Modifier.isStatic(field.getModifiers())) {
            if (true) {
              continue; //static fields already done in previous method
            }
            returnValue = field.get(null);
          } else if (classObject != null) {
            returnValue = field.get(classObject);
          } else {
            continue;
          }

          if (returnValue != null && !Utils.isWrapperType(returnValue.getClass()) && !Utils.isCollectionType(returnValue)) {
            // is Object
            System.out.println("returnValue = " + returnValue);
          }

          APIMethod.writeByteToSingleFileStream(("||" + className + "." + field.getName()).getBytes());
          APIMethod.writeByteToSingleFileStream((":::" + Utils.getStringRepresentation(returnValue) + "\n").getBytes());
          System.out.println("returnValue = " + returnValue);


        } catch (IllegalAccessException | IOException e) {
          e.printStackTrace();
        }
      }
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
      throw new RuntimeException();
    }
  }

  @Override
  public void onCreate() {
    super.onCreate();
    showNotification();
  }

  @Override
  public int onStartCommand(Intent intent, int flags, int startId) {
    showNotification();
    return super.onStartCommand(intent, flags, startId);
  }

  // required for newer versions of Android
  private void showNotification() {
    if (Build.VERSION.SDK_INT >= 26) {
      Log.i(TAG, "Showing notification");
      String CHANNEL_ID = "my_channel_01";
      NotificationChannel channel = new NotificationChannel(CHANNEL_ID, "Channel human readable title", NotificationManager.IMPORTANCE_DEFAULT);

      ((NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE)).createNotificationChannel(channel);

      Notification notification = new NotificationCompat.Builder(this, CHANNEL_ID).setContentTitle("APIHarvesting service started").setContentText("Collecting data").build();

      startForeground(1, notification);
    }
  }

  @Override
  public void onDestroy() {
    instance = null;
    super.onDestroy();
  }

  @Override
  protected void onHandleIntent(Intent workIntent) {
    instance = this;

    ThreadGroup group = new ThreadGroup("threadGroup");
    new Thread(group, () -> {
      // Load config
      Config.loadConfig(instance);

      cleanupFileDir();

      startLogging();
    }, "Harvesting Thread", 800000000).start();
  }

  // Deletes old record files
  private void cleanupFileDir() {
    File externalFilesDir = getExternalFilesDir(null);
    assert externalFilesDir != null;
    for (File file : externalFilesDir.listFiles()) {
      if (file.isFile()) {
        file.delete();
      }
    }
  }

  private void startLogging() {
    // Setup: "parse" parsed files, load predefines, try to invoke each method (recursion flattening)
    CustomHashMap<String, APIClass> methodsOfClasses;
    try {
      Parser parser = new Parser(this);
      CustomHashMap<String, APIClassInfo> methodsOfAllClassesParsed = parser.getMethodsOfAllClassesParsed();

      doStaticFields(methodsOfAllClassesParsed);

      APIMethod.KEEP_MEASUREMENTS = true;

      Predefines predefines = new Predefines(this);
      APIClassLoader classLoader = new APIClassLoader(this, predefines, parser);
      methodsOfClasses = classLoader.getMethodsOfClasses();
      stopLogging();
      if (true) {
        return;
      }
    } catch (IOException e) {
      e.printStackTrace();
      throw new RuntimeException("Init failed!");
    }
  }

  private void doStaticFields(@NonNull CustomHashMap<String, APIClassInfo> methodsOfAllClassesParsed) {
    APIMethod.getSingleFile(getApplicationContext().getExternalFilesDir(null));
    for (APIClassInfo apiClassInfo : methodsOfAllClassesParsed.values()) {
      String className = apiClassInfo.getClassName();
      try {
        Class<?> class_ = Class.forName(className);
        Field[] fields = class_.getDeclaredFields();
        for (Field field : fields) {

          field.setAccessible(true);
          Object returnValue;

          if (java.lang.reflect.Modifier.isStatic(field.getModifiers())) {
            returnValue = field.get(null);
          } else {
            continue;
          }

          if (returnValue != null && !Utils.isWrapperType(returnValue.getClass()) && !Utils.isCollectionType(returnValue)) {
            // is Object
            System.out.println("returnValue = " + returnValue);
          }

          APIMethod.writeByteToSingleFileStream(("||" + className + "." + field.getName()).getBytes());
          APIMethod.writeByteToSingleFileStream((":::" + Utils.getStringRepresentation(returnValue) + "\n").getBytes());
          System.out.println("returnValue = " + returnValue);
        }
      } catch (ClassNotFoundException e) {
      } catch (IllegalAccessException | IOException e) {
        e.printStackTrace();
      }
    }
  }

  public void stopLogging() {
  }
}